﻿
using BenchmarkDotNet.Attributes;
using BenchmarkDotNet.Running;

BenchmarkRunner.Run<VirtualMethodBenchmarks>();

public interface I1 {
	public int m(int x);
	public int f(int x);
}

public interface IDummy1 {
	public int m(int x);
	public int f(int x);
}

public interface IDummy2 {
	public int m(int x);
	public int f(int x);
}

public interface I2 {
	public int m(int x);
	public int f(int x);
}

public class A : I1, IDummy1, IDummy2, I2 {
	public int m(int x) {
		return x * 10 + 5;
	}

	[System.Runtime.CompilerServices.MethodImpl(System.Runtime.CompilerServices.MethodImplOptions.NoInlining)]
	public int mNonInlined(int x) {
		return x * 10 + 5;
	}

	public virtual int f(int x) {
		return x * 10 + 5;
	}
}

public class B : A {
	public new int m(int x) {
		return x * 10 + 6;
	}

	public override int f(int x) {
		return x * 10 + 6;
	}
}

public class C : A {
	public new int m(int x) {
		return x * 10 + 7;
	}

	public override int f(int x) {
		return x * 10 + 7;
	}
}

public class D : A {
	public new int m(int x) {
		return x * 10 + 8;
	}

	public sealed override int f(int x) {
		return x * 10 + 8;
	}
}

public sealed class E : A {
	public new int m(int x) {
		return x * 10 + 9;
	}

	public override int f(int x) {
		return x * 10 + 9;
	}
}

public class VirtualMethodBenchmarks {

	private A _a = new A();

	private B _b = new B();
	private C _c = new C();
	private D _d = new D();

	private E _e = new E();

	private A _aWithC;

	private I1 _i1A;
	private I1 _i1C;
	private I2 _i2A;

	private int _x;

	[GlobalSetup]
	public void GlobalSetup() {
		_aWithC = _c;

		_i1A = _a;
		_i1C = _c;
		_i2A = _a;

		_x = 3;
	}

	[Benchmark]
	public int NonVirtual() {
		return _a.m(3);
	}

	[Benchmark]
	public int NonVirtual_NoInlining() {
		return _a.mNonInlined(3);
	}

	[Benchmark]
	public int Virtual_AVar_AInstance() {
		return _a.f(3);
	}

	[Benchmark]
	public int Virtual_AVar_CInstance() {
		return _aWithC.f(3);
	}

	[Benchmark]
	public int Virtual_CVar_CInstance() {
		return _c.f(3);
	}

	[Benchmark]
	public int Virtual_DVar_DInstance() {
		return _d.f(3);
	}

	[Benchmark]
	public int Virtual_EVar_EInstance() {
		return _e.f(3);
	}

	[Benchmark]
	public int NonVirtual_TypeChoice_Switch_AVar_AInstance() {
		return _a switch {
			B b => b.m(3),
			C c => c.m(3),
			D d => d.m(3),
			A => _a.m(3),
		};
	}

	[Benchmark]
	public int NonVirtual_TypeChoice_Switch_AVar_CInstance() {
		return _aWithC switch {
			B b => b.m(3),
			C c => c.m(3),
			D d => d.m(3),
			A => _aWithC.m(3),
		};
	}

	[Benchmark]
	public int NonVirtual_TypeChoice_IfViaIs_AVar_AInstance() {
		if (_a is B b) {
			return b.m(3);
		} else if (_a is C c) {
			return c.m(3);
		} else if (_a is D d) {
			return d.m(3);
		} else {
			return _a.m(3);
		}
	}

	[Benchmark]
	public int NonVirtual_TypeChoice_IfViaIs_AVar_CInstance() {
		if (_aWithC is B b) {
			return b.m(3);
		} else if (_aWithC is C c) {
			return c.m(3);
		} else if (_aWithC is D d) {
			return d.m(3);
		} else {
			return _aWithC.m(3);
		}
	}

	[Benchmark]
	public int NonVirtual_TypeChoice_IfViaGetType_AVar_AInstance() {
		if (_a.GetType() == typeof(B)) {
			return ((B) _a).m(3);
		} else if (_a.GetType() == typeof(C)) {
			return ((C) _a).m(3);
		} else if (_a.GetType() == typeof(D)) {
			return ((D) _a).m(3);
		} else {
			return _a.m(3);
		}
	}

	[Benchmark]
	public int NonVirtual_TypeChoice_IfViaGetType_AVar_CInstance() {
		if (_aWithC.GetType() == typeof(B)) {
			return ((B) _aWithC).m(3);
		} else if (_aWithC.GetType() == typeof(C)) {
			return ((C) _aWithC).m(3);
		} else if (_aWithC.GetType() == typeof(D)) {
			return ((D) _aWithC).m(3);
		} else {
			return _aWithC.m(3);
		}
	}

	[Benchmark]
	public int DirectImpl_TypeChoice_IfViaGetType_AVar_AInstance() {
		if (_a.GetType() == typeof(B)) {
			return _x * 10 + 6;
		} else if (_a.GetType() == typeof(C)) {
			return _x * 10 + 7;
		} else if (_a.GetType() == typeof(D)) {
			return _x * 10 + 8;
		} else {
			return _x * 10 + 5;
		}
	}

	[Benchmark]
	public int DirectImpl_TypeChoice_IfViaGetType_aWithCVar_CInstance() {
		if (_aWithC.GetType() == typeof(B)) {
			return _x * 10 + 6;
		} else if (_aWithC.GetType() == typeof(C)) {
			return _x * 10 + 7;
		} else if (_aWithC.GetType() == typeof(D)) {
			return _x * 10 + 8;
		} else {
			return _x * 10 + 5;
		}
	}

	private Dictionary<Type, int> _typeToXMap = new() {
		{ typeof(A), 5},
		{ typeof(B), 6},
		{ typeof(C), 7},
		{ typeof(D), 8}
	};

	[Benchmark]
	public int DirectImpl_TypeChoice_Dictionary_AVar_AInstance() {
		return _x * 10 + _typeToXMap[_a.GetType()];
	}

	[Benchmark]
	public int DirectImpl_TypeChoice_Dictionary_AVar_CInstance() {
		return _x * 10 + _typeToXMap[_aWithC.GetType()];
	}

	[Benchmark]
	public int Constant_TypeChoice_Switch_AVar_AInstance() {
		return _a switch {
			B => 3 * 10 + 6,
			C => 3 * 10 + 7,
			D => 3 * 10 + 8,
			A => 3 * 10 + 5,
		};
	}

	[Benchmark]
	public int Constant_TypeChoice_IfViaGetType_AVar_AInstance() {
		if (_a.GetType() == typeof(B)) {
			return 3 * 10 + 6;
		} else if (_a.GetType() == typeof(C)) {
			return 3 * 10 + 7;
		} else if (_a.GetType() == typeof(D)) {
			return 3 * 10 + 8;
		} else {
			return 3 * 10 + 5;
		}
	}

	private Dictionary<Type, int> _typeToConstResultMap = new() {
		{ typeof(A), 3 * 10 + 5},
		{ typeof(B), 3 * 10 + 6},
		{ typeof(C), 3 * 10 + 7},
		{ typeof(D), 3 * 10 + 8}
	};

	[Benchmark]
	public int Constant_TypeChoice_Dictionary_AVar_AInstance() {
		return _typeToConstResultMap[_a.GetType()];
	}

	[Benchmark]
	public int Constant_TypeChoice_Dictionary_AVar_CInstance() {
		return _typeToConstResultMap[_aWithC.GetType()];
	}

	[Benchmark]
	public int NonVirtual_I1Var_AInstance() {
		return _i1A.m(3);
	}

	[Benchmark]
	public int NonVirtual_I2Var_AInstance() {
		return _i2A.m(3);
	}

	[Benchmark]
	public int Virtual_I1Var_AInstance() {
		return _i1A.f(3);
	}

	[Benchmark]
	public int Virtual_I2Var_AInstance() {
		return _i2A.f(3);
	}

	[Benchmark]
	public int Virtual_I1Var_CInstance() {
		return _i1C.f(3);
	}
}